/*
 *  USB Hanwang tablet support
 *
 *  Copyright (c) 2010 Xing Wei <weixing@hanwang.com.cn>
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
#include "hwesp_global.h"
#include <linux/types.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>

#define DRIVER_AUTHOR   "Houyun <Houyun@hanwang.com.cn>"
#define DRIVER_DESC     "USB Hanwang tablet driver"
#define DRIVER_LICENSE  "GPL"

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE(DRIVER_LICENSE);

static struct usb_interface *hwintf =NULL;
static int                g_ValidDev = 0;
static int                g_Probe = 0;


void hanwang_irq(struct urb *urb)
{
 //   printk(KERN_DEBUG "hanwang_irq ");

	struct hanwang *hanwang = urb->context;
	struct usb_device *dev = hanwang->usbdev;
    struct input_dev *input_dev = hanwang->dev;
    
    return irq(urb->status,urb,dev, dev->descriptor.idProduct,input_dev, hanwang->data);	
}

int hanwang_open(struct input_dev *dev)
{     
  //   printk(KERN_DEBUG "hanwang_open hwintf:%x\n",hwintf);
    if(!hwintf || (g_Probe == 0))
        return -ENODEV;
     struct hanwang *hanwang;     
	hanwang = input_get_drvdata(dev);
       if(!hanwang || !dev)
     {
           printk(KERN_DEBUG "device null");
         return -ENODEV;
    }

    hanwang->irq->dev = hanwang->usbdev;
	if (usb_submit_urb(hanwang->irq, GFP_KERNEL))
		return -EIO;
    g_ValidDev = true;
    if(hanwang->usbdev->descriptor.idProduct != 0x855c || hanwang->usbdev->descriptor.idProduct != 0x8563 || hanwang->usbdev->descriptor.idProduct != 0x8835)
     {
        if((g_ValidDev == false)  )
        {
         if( VarifyDev(hwintf) == 0)
           g_ValidDev = true;
         }
	}
    else 
        return 0;
if(!g_ValidDev)
         return -ENODEV;
	return 0;
}

 void hanwang_close(struct input_dev *dev)
{
    printk(KERN_DEBUG "hanwang_close\n");
	struct hanwang *hanwang = input_get_drvdata(dev);

	usb_kill_urb(hanwang->irq);
}

static int usb_set_report(struct usb_interface *intf, u8 type, u8 id,
			    void *buf, size_t size, unsigned int retries)
{
	struct usb_device *dev = interface_to_usbdev(intf);
	int retval;

	do {
		retval = usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
				USB_REQ_SET_REPORT,
				USB_TYPE_CLASS | USB_RECIP_INTERFACE,//0x21,
				(type << 8) + id,
				intf->altsetting[0].desc.bInterfaceNumber,
				buf, size, 1000);
	} while ((retval == -ETIMEDOUT || retval == -EPIPE) && --retries);

	return retval;
}
int usb_send_device_data(struct usb_interface *intf, int report_id, int length, unsigned char *user_buffer)
{
	unsigned char *rep_data;
	int error = -ENOMEM, limit = 0;
	rep_data = kzalloc(length + 1, GFP_KERNEL);  
	
	if (!rep_data)
		return error;

	if (memmove(rep_data, user_buffer, length) == 0) 	
	{
		kfree(rep_data);
		return error;
	}
	do {
		error = usb_set_report(intf, HWESP_HID_FEATURE_REPORT,
		                         report_id, rep_data, length, 20);
	} while ((error < 0) && limit++ < HWESP_MSG_RETRIES);

	kfree(rep_data);
	return error < 0 ? error : 0;
}

static struct hanwang_features features = {0x0,"HanWang ESPMonitor",0,
         0,0x0,0x0, 0x0,0x0,0};
static int hanwang_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
struct usb_device *dev = interface_to_usbdev(intf);
	struct usb_endpoint_descriptor *endpoint;
	struct hanwang *hanwang;
	struct input_dev *input_dev;
	int error;
	int i;
	hanwang = kzalloc(sizeof(struct hanwang), GFP_KERNEL);
	input_dev = input_allocate_device();
	if (!hanwang || !input_dev) {
		error = -ENOMEM;
		goto fail1;
	}

    if (!get_features(dev->descriptor.idProduct, &features)) {
		error = -ENXIO;
		goto fail1;
	}
    hanwang->features = &features;
 //   printk(KERN_DEBUG "features:%x %s %d\n",hanwang->features->pid,hanwang->features->name,hanwang->features->pkg_len);

    hanwang->data = usb_alloc_coherent(dev,  hanwang->features->pkg_len,  GFP_KERNEL, &hanwang->data_dma);
	if (!hanwang->data) {
		error = -ENOMEM;
		goto fail1;
	}

	hanwang->irq = usb_alloc_urb(0, GFP_KERNEL);
	if (!hanwang->irq) {
		error = -ENOMEM;
		goto fail2;
	}

	hanwang->usbdev = dev;
	hanwang->dev = input_dev;
	hanwang->current_id = 0;

	usb_make_path(dev, hanwang->phys, sizeof(hanwang->phys));
	strlcat(hanwang->phys, "/input0", sizeof(hanwang->phys));

	strlcpy(hanwang->name, hanwang->features->name, sizeof(hanwang->name));
	input_dev->name = hanwang->name;
	input_dev->phys = hanwang->phys;
	usb_to_input_id(dev, &input_dev->id);
	input_dev->dev.parent = &intf->dev;

	input_set_drvdata(input_dev, hanwang);
	input_dev->open = hanwang_open;
	input_dev->close = hanwang_close;

    error = probe(input_dev,dev->descriptor.idProduct,input_dev->evbit,input_dev->absbit, input_dev->keybit);
    if(error != 0)
      goto fail3;

	endpoint = &intf->cur_altsetting->endpoint[0].desc;
	usb_fill_int_urb(hanwang->irq, dev,
			usb_rcvintpipe(dev, endpoint->bEndpointAddress),
			hanwang->data, hanwang->features->pkg_len,
			hanwang_irq, hanwang, endpoint->bInterval);
	hanwang->irq->transfer_dma = hanwang->data_dma;
	hanwang->irq->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

	error = input_register_device(hanwang->dev);
	if (error)
		goto fail3;

	usb_set_intfdata(intf, hanwang);
    hwintf = intf;
    g_Probe = 1;
    unsigned char bSwitchWtb[3] ={0x02,0x02,0x02};
    usb_send_device_data(intf, 0x02, 2, bSwitchWtb);
	return 0;

 fail3:	usb_free_urb(hanwang->irq);
 fail2:	usb_free_coherent(dev, hanwang->features->pkg_len,
			hanwang->data, hanwang->data_dma);
 //fail2:	usb_buffer_free(dev, hanwang->features->pkg_len,
 //                          hanwang->data, hanwang->data_dma);
  fail1:	input_free_device(input_dev);
	kfree(hanwang);
	return error;
}

static void hanwang_disconnect(struct usb_interface *intf)
{
   g_ValidDev = false;
//   g_ResOk = 0;
	struct hanwang *hanwang = usb_get_intfdata(intf);

	input_unregister_device(hanwang->dev);
	usb_free_urb(hanwang->irq);
        usb_free_coherent(interface_to_usbdev(intf),
			hanwang->features->pkg_len, hanwang->data,
			hanwang->data_dma);

	kfree(hanwang);
	usb_set_intfdata(intf, NULL);
    hwintf = NULL;
    g_Probe = 0;
}

MODULE_DEVICE_TABLE(usb, hanwang_ids);

static struct usb_driver hanwang_driver = {
        .name		= "esp2210",
	.probe		= hanwang_probe,
        .disconnect	= hanwang_disconnect,
	.id_table	= hanwang_ids,
};

static int __init hanwang_init(void)
{
	return usb_register(&hanwang_driver);
}

static void __exit hanwang_exit(void)
{  
	usb_deregister(&hanwang_driver);
}

module_init(hanwang_init);
module_exit(hanwang_exit);
